<?php

namespace Zsd\Db;

use ZendServer\Set,
ZendServer\Log\Log,
ZendServer\FS\FS,
Zend\Json\Json,
Zend\Db\TableGateway\TableGateway,
ZendServer\Exception;
use Zend\Db\Sql\Select;

class TasksMapper {
	
	const DUMMY_NODE_ID = -1;
	const DUMMY_TASK_ID = -1;
	const DUMMY_AUDIT_ID = -1;
	
	//////////////////////////////////////////////////////////////////
	// ZSD coommands
	//////////////////////////////////////////////////////////////////
	
	// restart Zend Server (all daemons + apache)
	const COMMAND_RESTART_SERVER           = 0;

	// Remove the current node from the cluster and move to "singleserver" mode
	const COMMAND_DISCONNECT_FROM_CLUSTER  = 1;
	
	// Copy the blueprint from the database to the INI files
	const COMMAND_SAVE_AND_APPLY_BLUEPRINT = 2;
	
	// join to cluster
	const COMMAND_JOIN_CLUSTER             = 3;
	
	// enable an extension. The extension name is defined
	// inside the extra-info (json)
	const COMMAND_ENABLE_EXTENSION         = 4;
	
	// disable an extension. The extension name is defined
	// inside the extra-info (json)
	const COMMAND_DISABLE_EXTENSION        = 5;
	
	// Disable the current server
	const COMMAND_DISABLE_SERVER           = 6;
	
	// Enable the current server
	const COMMAND_ENABLE_SERVER            = 7;
	
	// An internal task that is passed by the "removed" server to the other servers in the
	// cluster notifying about its removal
	const COMMAND_CLUSTER_MEMBER_REMOVED   = 8;
	
	// applies blueprint data to FS against a certain node
	const COMMAND_APPLY_BLUEPRINT          = 12;

	// Force remove a server. The node id is defined
	// inside the extra-info (json)
	const COMMAND_FORCE_DISCONNECT_FROM_CLUSTER = 13;	
	
	const COMMAND_RESTART_PHP = 15;// Restart the webserver	
	const COMMAND_RESTART_MONITOR_NODE = 16;// Restart Monitor Node	
	const COMMAND_RESTART_SCD = 17;// Restart Session Clustering	
	const COMMAND_RESTART_JQD = 18;// Restart Job Queue Daemon	
	const COMMAND_RESTART_ZDD = 19;// Restart Deployment Daemon	
	const COMMAND_RESTART_JB = 29;// Restart JB
	
	// Monitor rules were updated in the database
	const COMMAND_MONITOR_RULES_UPDATED = 20;
	
	const COMMAND_CLEAR_OPTIMIZER_PLUS_CACHE = 21;
	const COMMAND_CLEAR_DATACACHE_DISK_CACHE = 22;
	const COMMAND_CLEAR_DATACACHE_SHM_CACHE = 23;
	const COMMAND_CLEAR_PAGECACHE_CACHE = 24;
	
	const COMMAND_PAGECACHE_RULES_UPDATED = 26;// Page Cache rules were updated in the database
	
	const COMMAND_GET_PHPINFO = 28;	
	
	const COMMAND_APPS_LIST_UPDATED = 39;
	
	// applies blueprint data to FS against a certain node - including extension data
	const COMMAND_APPLY_BLUEPRINT_WITH_EXTENSIONS          = 40;
	
	const COMMAND_LICENSE_UPDATED = 41;
	
	const COMMAND_RELOAD_CONFIGURATION = 42;
	
	const RESTART_TYPE_FULL = 0;
	const RESTART_TYPE_SELECTIVE = 1;
	
	protected $daemonsToCommands = array(
			'jb' 			=> self::COMMAND_RESTART_JB,
			'jqd' 			=> self::COMMAND_RESTART_JQD,
			'monitor_node'	=> self::COMMAND_RESTART_MONITOR_NODE,
			'scd' 			=> self::COMMAND_RESTART_SCD,
			'zdd' 			=> self::COMMAND_RESTART_ZDD,			
			);
	
	/**
	 * @var \Zend\Db\TableGateway\TableGateway
	 */
	private $tableGateway;
	
	public function __construct(TableGateway $tableGateway = null) {
		$this->tableGateway = $tableGateway;
	}
	
	/**
	 * @param integer $serverId
	 * @param integer $taskId
	 * @param integer $auditId
	 * @param mixed $extraData
	 * @return integer
	 */
	public function insertTask($serverId, $taskId, $auditId = self::DUMMY_TASK_ID, $extraData = array()) {
		if (is_object($extraData) || is_array($extraData)) {
			$extraData = Json::encode($extraData);
		}
		$this->tableGateway->insert(
				array(	'NODE_ID'	=> $serverId,
						'TASK_ID' 	=> $taskId,
						'AUDIT_ID' => $auditId,
						'EXTRA_DATA' => $extraData,
				));
		Log::debug("Insert task s:{$serverId}, t:{$taskId}, a:{$auditId}, ex:{$extraData}");
		return $this->tableGateway->getLastInsertValue();
	}
	
	/**
	 * @param array $serverIds
	 * @param integer $taskId
	 * @param integer $auditId
	 * @param mixed $extraData
	 */
	public function insertTasksServers($serverIds, $taskId, $auditId = self::DUMMY_TASK_ID, $extraData = array()) {
		$dbConnection = $this->getTableGateway()->getAdapter()->getDriver()->getConnection(); /* @var $dbConnection \Zend\Db\Adapter\Driver\ConnectionInterface */
		try {
			$dbConnection->beginTransaction();
			foreach($serverIds as $serverId) {
				$this->insertTask($serverId, $taskId, $auditId, $extraData);
			}
			$dbConnection->commit();
		} catch (Exception $ex) {
			$dbConnection->rollback();
			Log::err("Cannot send task {$taskId}: {$ex->getMessage()}");
		}
	}
	
	/**
	 * @throws Exception
	 * @return string
	 */
	public function getTaskDaemonName($daemonName) {
		if (!isset($this->daemonsToCommands[$daemonName])) {
			throw new Exception("Failed to restart- unknown daemon '{$daemonName}'");
		}
		return $this->daemonsToCommands[$daemonName];
	}
	
	/**
	 * @param integer $serverId
	 */
	public function removeServerTasks($serverId) {
		Log::debug("removing server '$serverId' current tasks");
		$this->tableGateway->delete(array('NODE_ID' => $serverId));
	}
	
	/**
	 * @param array $directives - array of directives consisting of keys=>values
	 * @param integer $auditId - the auditId to be inserted
	 */
	public function setDirectives(array $directives, $auditId=self::DUMMY_TASK_ID) {
		Log::info("set directives {" . implode(',', array_keys($directives)) . '}');	
		
		$directivesToSave = array();
		foreach ($directives as $name => $value) {
			$directivesToSave[] = array('name' => $name, 'value' => strval($value));
		}
		$this->tableGateway->insert(array('NODE_ID' => self::DUMMY_NODE_ID, 'TASK_ID' => self::COMMAND_SAVE_AND_APPLY_BLUEPRINT, 'EXTRA_DATA'=>json_encode($directivesToSave), 'AUDIT_ID' => $auditId));
	}
	
	/**
	 * @param array $serverId
	 */
	public function applyBlueprintToServers(array $serverIds) {
		foreach ($serverIds as $serverId) {
			$this->applyBlueprint($serverId);
		}
	}
	
	/**
	 * @param integer $serverId
	 */
	public function applyBlueprint($serverId) {
		Log::info("apply blueprint for server " . $serverId);
	
		$this->tableGateway->insert(array('NODE_ID' => $serverId, 'TASK_ID' => self::COMMAND_APPLY_BLUEPRINT_WITH_EXTENSIONS));
	}
	
	/**
	 * @param array $directives - array of directives consisting of keys=>values. keys might not include the zend_gui. prefix
	 * @param integer $auditId - the auditId to be inserted
	 */
	public function setGuiDirectives(array $directives, $auditId=self::DUMMY_TASK_ID) {
		$directivesToSave = array();
		$directivesPrefix = \Application\Module::INI_PREFIX . '.';
		foreach($directives as $name => $value) {
			$name = preg_replace("/^({$directivesPrefix}){0,1}(.*)/", "{$directivesPrefix}\${2}", $name); // making sure we have directive starting with zend_gui. whether the prefix was passed or not
			$directivesToSave[] = array('name' => $name, 'value' => strval($value));
		}
		
		Log::info("Updating GUI directives with the following changes: " . print_r($directivesToSave, true));

		$this->tableGateway->insert(array('NODE_ID' => self::DUMMY_NODE_ID, 'TASK_ID' => self::COMMAND_SAVE_AND_APPLY_BLUEPRINT, 'EXTRA_DATA'=>json_encode($directivesToSave), 'AUDIT_ID' => $auditId));
	}

	/**
	 * @param array $servers
	 * @return array
	 */
	function tasksPerServerComplete(array $servers) {
		if (count($servers) > 0) {
			$resultSet = $this->tableGateway->select('NODE_ID IN (' . implode(',', $servers). ')');
			$resultSet = $resultSet->toArray();
			$result = array_combine($servers, array_fill(0, count($servers), true));
			foreach ($resultSet as $taskRow) {
				$result[$taskRow['NODE_ID']] = false;
			}
			return $result ? $result : array();
		}
		return array();
	}
	
	/**
	 * @return \Zend\Db\ResultSet\ResultSet
	 */
	public function findAllTasks() {
		return $this->getTableGateway()->select();
	}
	
	/**
	 * @return \Zend\Db\ResultSet\ResultSet
	 */
	public function findAllTasksOfConnectedServers() {
		$select = new Select();
		$select->join('ZSD_NODES', 'ZSD_NODES.NODE_ID = ZSD_TASKS.NODE_ID AND ZSD_NODES.IS_DELETED = 0', array());
		$select->from($this->getTableGateway()->getTable());
		return $this->getTableGateway()->selectWith($select);
	}
	
	/**
	 * 
	 * @param array $servers
	 * @param array $tasks
	 * @return boolean
	 */
	public function tasksComplete($servers = array(), $tasks = array()) {
	    // @todo: Preferable to use COUNT(*)
	    	  
	    if (count($servers) > 0) {
		    $resultSet = $this->tableGateway->select('NODE_ID IN (' . implode(',', $servers). ')');
	    }
	    if (count($tasks) > 0) {
		    $resultSet = $this->tableGateway->select('TASK_ID IN (' . implode(',', $tasks). ')');
	    }
	    $resultSet = $this->tableGateway->select();
	    
	    /* @var $resultSet \Zend\Db\ResultSet\ResultSet */
	    return (! $resultSet || count($resultSet->toArray()) > 0) ? false : true;
	}
	
	public function waitForTasksComplete($servers = array(), $tasks = array(), $sleep=3) {
		$max_time = ini_get('max_execution_time');
		$initTime = $_SERVER["REQUEST_TIME"];
		
		while (($timePassed = round(time() - $initTime)) < $max_time - ($sleep+1)) { // will stop the check, once we're about to reach the max_execution_time limit
			if ($this->tasksComplete($servers, $tasks)) {
				Log::info("All tasks were completed after {$timePassed} seconds)");
				return true;
			}

			Log::debug("waiting for TasksComplete ({$timePassed} seconds passed)");
			sleep($sleep);
		}
		
		throw new Exception("tasks have not been completed!");
	}

	/**
	 * @param array $ids
	 */
	public function syncMonitorRulesChanges(array $ids) {
		Log::debug(__FUNCTION__ . " on servers ".implode(',', $ids));
	
		foreach ($ids as $id) {
			$this->tableGateway->insert(array('NODE_ID'	=> $id,	'TASK_ID' => self::COMMAND_MONITOR_RULES_UPDATED ));
		}
	}	
	
	/**
	 * @param array $ids
	 */
	public function syncPageCacheRulesChanges(array $ids) {
		Log::debug(__FUNCTION__ . " on servers ".implode(',', $ids));
	
		foreach ($ids as $id) {
			$this->tableGateway->insert(array('NODE_ID'	=> $id,	'TASK_ID' => self::COMMAND_PAGECACHE_RULES_UPDATED ));
		}
	}
	
	/**
	 * @return \Zend\Db\TableGateway\TableGateway $tableGateway
	 */
	public function getTableGateway() {
		return $this->tableGateway;
	}

	/**
	 * @param \Zend\Db\TableGateway\TableGateway $tableGateway
	 * @return \Servers\Db\Mapper
	 */
	public function setTableGateway($tableGateway) {
		$this->tableGateway = $tableGateway;
		return $this;
	}

}
