<?php

namespace Configuration;
use ZendServer\Log\Log;
use ZendServer\Exception;
use ZendServer\Validator\ErrorReporting;
use ZendServer\Validator\NullValidator;

use Zend\Validator\Between;

use Zend\Validator\ValidatorChain;

use ZendServer\Validator\Integer;

use ZendServer\Validator\Boolean;

use Zend\Validator\InArray;

use Zend\Validator\Regex;


use Zend\Json\Json,
ZendServer\Set;
use ZendServer\Validator\DirectiveStringValidator;
use ZendServer\Validator\PhpExpression;

class DdMapper {
	const DIRECTIVE_TYPE_STRING = 1;
    const DIRECTIVE_TYPE_BOOLEAN = 2;
    const DIRECTIVE_TYPE_INTEGER = 4;
    /**
	 * 
	 * @var stdClass
	 */
	public $data;
	
	/**
	 * 
	 * @param string $zeMapFileContent
	 */
	public function __construct($zeMapFileContent) {
		$this->data = Json::decode($zeMapFileContent, Json::TYPE_ARRAY);
	}
	
	/**
	 * @param \Configuration\DirectiveContainer $directive
	 * @return array
	 * @throws \ZendServer\Exception
	 */
	public function addDirectiveData (\Configuration\DirectiveContainer $directive) {
		$directive = $directive->toArray();
		
		if (! isset($directive['NAME'])) {
			throw new Exception('Provided directive array should contain the \'NAME\' field');
		}
	
		$directive['DAEMON'] ? $componentName = $directive['DAEMON'] : $componentName = $directive['EXTENSION']; // we check also component, as some directives appear both under extension and daemon
		$directiveData = $this->getDirectiveRow($directive['NAME'], $componentName);
		return array_merge($directive, (array) $directiveData);
	}
	
	/**
	 * 
	 * @param array|Set $directives
	 * @throws \ZendServer\Exception
	 * @return \ZendServer\Set
	 */
	public function addDirectivesData($directives) {
		$detailedDirectives = array();
		foreach ($directives as $directive) {
			if (!$directive instanceOf \Configuration\DirectiveContainer) {
				throw new Exception('Provided directives array should contain the DirectiveContainer objects');
			}
			try {
				$directiveData = $this->addDirectiveData($directive);
				$directiveContext = $directive->getContext();
				$directiveContextName = $directive->getContextName();
			} catch (Exception $ex) {
				Log::notice("Directive {$directive->getName()} is missing data and will not be displayed");
				continue;
			}
			$detailedDirectives[$directive->getName()] = $directiveData;
		}
		
		return new Set($detailedDirectives, '\Configuration\DirectiveContainer');
	}

	/**
	 *
	 * @param array|Set $directives
	 * @throws \ZendServer\Exception
	 * @return \ZendServer\Set
	 */
	public function addDirectivesMetadataOnly($directives, $ext) {
		$detailedDirectives = array();
		foreach ($directives as $directive) {
			$detailedDirectives[$directive] = $this->getDirectiveRow($directive, $ext);
			$detailedDirectives[$directive]['NAME'] = $directive;
		}

		return new Set($detailedDirectives, '\Configuration\DirectiveContainer');
	}
	
	
	/**
	 * 
	 * @param array $extension
	 * @throws \ZendServer\Exception
	 * @return array
	 */
	public function addExtensionData(\Configuration\ExtensionContainer $extension) {
		$extension = $extension->toArray();
		
		if (! isset($extension['NAME'])) {
			throw new Exception('Provided extension array should contain the \'NAME\' field');
		}

		$extName = $extension['NAME'];
		if (!isset($this->data[$extName])) {
			Log::notice("could not find extension {$extName} in Zend Server metadata - user defined?");
			return $extension;
		}

		$dataArray = (array) $this->data[$extName];
		unset($dataArray['directives']);
		return array_merge($extension, $dataArray);		
	}
	
	/**
	 *
	 * @param array $daemon
	 * @throws \ZendServer\Exception
	 * @return array
	 */
	public function addDaemonData(\Configuration\DaemonContainer $daemon) {
		$daemon = $daemon->toArray();
		if (! isset($daemon['name'])) {
			throw new Exception('Provided extension array should contain the \'name\' field');
		}
	
		$daemonName = $daemon['name'];
		if (!isset($this->data[$daemonName])) {
			Log::notice("could not find extension {$daemonName} in Zend Server metadata - user defined?");
			return $daemon;
		}
	
		$dataArray = (array) $this->data[$daemonName];
		unset($dataArray['directives']);
		return array_merge($daemon, $dataArray);
	}
	
	/**
	 *
	 * @param array|Set $daemons
	 * @throws \ZendServer\Exception
	 * @return \ZendServer\Set
	 */
	public function addDaemonsData ($daemons) {
		$detailedDaemons = array();
		foreach ($daemons as $idx => $daemon) {
			if (!$daemon instanceOf \Configuration\DaemonContainer) {
				throw new Exception('Provided daemons array should contain the DaemonContainer objects');
			}
			$detailedDaemons[] = $this->addDaemonData($daemon);
		}
	
		return new Set($detailedDaemons, '\Configuration\DaemonContainer');
	}
	
	/**
	 * 
	 * @param array $extensions
	 * @throws \ZendServer\Exception
	 * @return \ZendServer\Set
	 */
	public function addExtensionsData (array $extensions) {
		$detailedExtensions = array();
		foreach ($extensions as $idx => $extension) {
			if (!$extension instanceOf \Configuration\ExtensionContainer) {
				throw new Exception('Provided extensions array should contain the ExtensionContainer objects');
			}
			$detailedExtensions[$idx] = $this->addExtensionData($extension);
		}
		
		return new Set($detailedExtensions, '\Configuration\ExtensionContainer');
	}
	
	/**
	 * 
	 * @param \Configuration\ExtensionContainer or \Configuration\DaemonContainer $extension
	 * @param string $filter
	 * @return bool
	 */
	public function matchExtensionStrings($extension, $filter) {
		if (! isset($this->data[$extension->getName()])) {
			Log::notice("could not find extension {$extension->getName()} in Zend Server metadata - user defined?");
			return false;
		}

		if (! $filter) {
			return true;
		}
		
		// match extension name/shortDescription/logDescription
		if(	strstr(strtolower($extension->getName()), $filter) ||
			strstr(strtolower($extension->getShortDescription()), $filter) ||
			strstr(strtolower($extension->getLongDescription()), $filter)) {
			return true;
		}
		
		return false;
	}

	/**
	 *
	 * @param \Configuration\ExtensionContainer or \Configuration\DaemonContainer $extension
	 * @param string $filter
	 * @return bool
	 */
	public function matchExtensionDirectives($extension, $filter) {
		if (! isset($this->data[$extension->getName()])) {
			Log::notice("could not find extension {$extension->getName()} in Zend Server metadata - user defined?");
			return false;
		}
	
		// match the directive/directiveDescription
		$extData = $this->data[$extension->getName()];			
		$matchedDirectives = array();
		foreach ($extData['directives'] as $name => $directive) {
			if (	strstr(strtolower($directive["shortDescription"]), $filter) ||
					strstr(strtolower($name), $filter)) {
				$matchedDirectives[] = $name;
			}
		}
	
		return $matchedDirectives;
	}
		
	/**
	 * 
	 * @param \Configuration\DirectiveContainer $directive
	 * @param string $filter
	 */
	public function matchDirective(\Configuration\DirectiveContainer $directive, $filter) {
		if (! $filter) {
			return true;
		}
		if (strstr(strtolower($directive->getName()), $filter)) {
			return true;
		}
		if (isset($this->data[$directive->getExtension()])) {
			$directivesArray = $this->data[$directive->getExtension()]['directives'];
			if (isset($directivesArray[$directive->getName()])) {
				if (strstr(strtolower($directivesArray[$directive->getName()]['shortDescription']), $filter)) {
						return true;
					
				}
			}
		}
		return false;
	}
	
	public function getDummyExtensions() {
		$dummyExtensions = array();
		
		foreach ($this->data as $extName => $data) {
			if (isset($data['dummy']) && $data['dummy']) {
				$dummyExtensions[] = $extName;
			}			
		}
		
		return $dummyExtensions;
	}
		
	/**
	 * @param string $directiveName
	 * @return \Zend\Validator\Validator
	 */
	public function directiveValidator($directiveName) {
		$directiveRow = $this->getDirectiveRow($directiveName);
		if (!$directiveRow) {
			Log::debug("The directive $directiveName wasn't found in the map file, so we got an unrecognized directive");
			return new NullValidator();
		}
		
		$validation = $directiveRow['validation'];		

        /// special cases
        if ($directiveName == 'error_reporting') {
            $validator = new ErrorReporting();
        } elseif ($validation['html']) {
			$validator = new Regex('#^[[:print:]]*$#'); /// allow any graphical characters
		} elseif ($validation['regex']) {
			$validator = new Regex($validation['regex']);
		} elseif ($validation['listValues']) {
			$validator = new InArray();
			$validator->setHaystack(array_keys((array) unserialize($validation['listValues']))); // the keys represent the values
			$validator->setStrict(InArray::COMPARE_NOT_STRICT_AND_PREVENT_STR_TO_INT_VULNERABILITY); // as serialize() and unserialize() causes array keys such as ["1"] to become [1], we loose the strict check
		} elseif ($validation['minValue'] || $validation['maxValue']) {
			$options = array('inclusive' => true);
			
			if ($validation['minValue']) {
				$options['min'] = $validation['minValue'];
			}
			
			if ($validation['maxValue']) {
				$options['max'] = $validation['maxValue'];
			}
			$validator = new ValidatorChain();
			// if it's between validator we should add aslo the numeric validator, since mixed values like abc5 returns isValid() true between 0 and 1000
			$validator->addValidator(new Between($options));
			$validator->addValidator(new Integer());
		} elseif ($directiveRow['type'] == self::DIRECTIVE_TYPE_INTEGER) { // integer
			$validator = new Integer();
		} elseif ($directiveRow['type'] == self::DIRECTIVE_TYPE_BOOLEAN) { // boolean
			$validator = new Boolean();
		} else {
			$validator = new DirectiveStringValidator(); // string validator
		}
		
		Log::debug("Validator for {$directiveName} is " . get_class($validator));
		return $validator;
	}
	

	/**
	 * @param string $directiveName
	 * @param string $componentName
	 * @return array
	 */
	private function getDirectiveRow($directiveName, $componentName='') { // @todo - temporary workaround till configurationValidateDirectives() will receive extension	
		foreach($this->data as $component => $data) {
			if (isset($data['directives'][$directiveName]) && ($componentName === '' || $componentName === $component)) {
				return $data['directives'][$directiveName];
			}
		}
		return array();
	}
}
