<?php

namespace Configuration\License;

use Zend\EventManager\EventManagerInterface;

use Zend\EventManager\ListenerAggregateInterface;

use Users\Db\Mapper;

use Zsd\Db\TasksMapper as tasksMapper;

use Zend\Permissions\Acl\Acl;

use ZendServer\Log\Log;

use ZendServer\Exception;

use Configuration\License\LicenseChangeAnalyzer;

use Zend\EventManager\Event;

use \Configuration\License\LicenseChangeContainer;

use Application\Module;

use Snapshots\Mapper\Profile;
use Configuration\MapperDirectives;
use Configuration\Controller\ZendMonitorController;

class ChangeListener implements ListenerAggregateInterface {
	
	/**
	 * @var LicenseChangeAnalyzer
	 */
	private $licenseChangeAnalyzer;
	
	/**
	 * @var Acl
	 */
	private $acl;
	
	/**
	 * @var Mapper
	 */
	private $usersMapper;
	
	/**
	 * @var tasksMapper
	 */
	private $tasksMapper;	
	
	/**
	 * @var Profile
	 */
	private $profileMapper;
	/**
	 * @var array
	 */
	private $listeners = array();
	
	/**
	 * @var MapperDirectives
	 */
	private $directivesMapper;
	
	/**
	 * @param Event $event
	 * @return boolean
	 */
	public function licenseChanged(Event $event) {
		$newDirectives = $event->getParam('newDirectives');
		$currentDirectives = $event->getParam('directives');
		/// changed the serial_number?
		if (! isset($newDirectives['zend.serial_number'])) {
			return false;
		}
		
		$this->directivesMapper = $event->getTarget();
		
		$newLicenseSerial = $newDirectives['zend.serial_number'];
		$newLicenseUser = $newDirectives['zend.user_name'];

		try {
			if (isset($currentDirectives['zend.serial_number'])) {
				$currentLicenseSerial = $currentDirectives['zend.serial_number'];
				$currentLicenseName = $currentDirectives['zend.user_name'];
			} else {
				$currentLicenseSerial = $currentLicenseName = null;
			}
			
			$licenseChangeContainer = $this->licenseChangeAnalyzer->analyzeLicenseChange($newLicenseSerial, $newLicenseUser, $currentLicenseSerial, $currentLicenseName);
			$this->checkEvaluationChange($licenseChangeContainer->getEvaluationChange());
			$this->checkEditionChange($licenseChangeContainer->getEditionChange(), $licenseChangeContainer->getNewEdition());
			$this->checkProfileCompliance($licenseChangeContainer);			
		} catch (Exception $ex) {
			return Log::warn("License change analysis failed: {$ex->getMessage()}");
		}

		// if all went well, then it is time to set trialExtended=1, as not to present the 21day option ever again
		$this->tasksMapper->setGuiDirectives(array('trialExtended'=>1)); 
	}
	
	/**
	 * @param LicenseChangeContainer $licenseChangeContainer
	 */
	protected function checkProfileCompliance($licenseChangeContainer) {
		$newEdition = $licenseChangeContainer->getNewEdition();
		/// check what's the current profile
		if (strtolower(Module::config('package', 'serverProfile')) != 'Development') {
			if (strtolower($newEdition) == 'developer') {
				/// change to developer profile
				$this->getProfileMapper()->activateProfile('developmentDirectives');
				// TODO create a new SystemBoot snapshot
			}
		}
	}
	
	/**
	 * @param integer $evaluationChange
	 */
	protected function checkEvaluationChange($evaluationChange) {
		if ($evaluationChange === LicenseChangeContainer::EVALUATION_TO_EVALUATION) {
			return Log::info('Evaluation license renewed');
		} 
		
		if ($evaluationChange === LicenseChangeContainer::EVALUATION_TO_COMMERCIAL) {
			return Log::info('Upgrading from evaluation license to commercial license');
		}

		if ($evaluationChange === LicenseChangeContainer::COMMERCIAL_TO_EVALUATION) {
			return Log::info('Downgrading from commercial license to evaluation license');
		}

		if ($evaluationChange === LicenseChangeContainer::COMMERCIAL_TO_COMMERCIAL) {
			return Log::info('Upgrading from commercial license to commercial license');
		}
					
		throw new Exception("Unknown evaluation change '$evaluationChange' detected!");
	}
		
	/**
	 * @param integer $editionChange
	 */	
	protected function checkEditionChange($editionChange, $newEdition) {
		if ($editionChange === LicenseChangeContainer::EDITION_NO_CHANGE) { // no change
			return Log::info('New license is of the same edition, no changes');
		}
		
		if ($editionChange === LicenseChangeContainer::EDITION_UPGRADE) { // upgrading
			return Log::info('License is upgrading');
		}
		
		Log::info('License is downgrading');// downgrading
		$newEdition = strtoupper($newEdition);
		$this->editionChange = LicenseChangeContainer::EDITION_DOWNGRADE;
		$editionRole = "edition:{$newEdition}";
		
		if (! $this->acl->isAllowed($editionRole, 'data:useMultipleUsers')) {
			$this->usersMapper->deleteAllButAdmin();// clean up all users but the admin
			Log::info('Remove gui access users that are not administrators');
		}
		
		if (! $this->acl->isAllowed($editionRole, 'data:collectEventsCodeTrace')) {
			$this->directivesMapper->setDirectives(array('zend_monitor.event_tracing_mode' => ZendMonitorController::EVENT_TRACING_MODE_OFF));
			Log::info('Set zend_monitor.tracing_mode to \'Off\'');
		}
		
	}

	/**
	 * @return Profile
	 */
	public function getProfileMapper() {
		return $this->profileMapper;
	}

	/**
	 * @param \Snapshots\Mapper\Profile $profileMapper
	 */
	public function setProfileMapper($profileMapper) {
		$this->profileMapper = $profileMapper;
	}

	/**
	 * @param \Configuration\License\LicenseChangeAnalyzer $licenseChangeAnalyzer
	 * @return ChangeListener
	 */
	public function setlicenseChangeAnalyzer($licenseChangeAnalyzer) {
		$this->licenseChangeAnalyzer = $licenseChangeAnalyzer;
		return $this;
	}

	/**
	 * @param Acl $acl
	 * @return ChangeListener
	 */
	public function setAcl(Acl $acl) {
		$this->acl = $acl;
		return $this;
	}

	/**
	 * @param \Users\Db\Mapper $usersMapper
	 * @return ChangeListener
	 */
	public function setUsersMapper($usersMapper) {
		$this->usersMapper = $usersMapper;
		return $this;
	}

	/**
	 * @param \Zsd\Db\TasksMapper $tasksMapper
	 * @return ChangeListener
	 */
	public function setTasksMapper($tasksMapper) {
		$this->tasksMapper = $tasksMapper;
		return $this;
	}
	
	public function attach(EventManagerInterface $events) {
		$this->listeners[] = $events->attach('setDirectives', array($this, 'licenseChanged'));
	}
	
	public function detach(EventManagerInterface $events) {
		$events->detach(current($this->listeners));
	}	
}
