<?php

namespace JobQueue\Controller;

use Acl\License\Exception as LicenseException;

use JobQueue\Model\JobsSet;
use JobQueue\Model\RulesSet;

use Audit\Db\ProgressMapper;

use Audit\Db\Mapper;

use ZendServer\Mvc\Controller\WebAPIActionController,
ZendServer\Exception;

use Zend\Mvc\Controller\ActionController,
	Application\Module,
	WebAPI,
	Zend\Stdlib,
	ZendServer\Log\Log,
	Zend\Validator,
	ZendServer,
	ZendServer\Set;

use JobQueue\Filter\Filter,
JobQueue\Filter\Translator,
JobQueue\Filter\Dictionary;


use Zend\View\Model\ViewModel;

class WebAPIController extends WebAPIActionController
{

	/**
	 * @var Dictionary
	 */
	protected $dictionary;
	
	public function jobqueueStatisticsAction() {
		$this->isMethodGet();
		$statistics = $this->getJobqueueMapper()->getStatistics();
		return array('statistics' => $statistics);
	}
	
	public function jobqueueRequeueJobsAction() {		
		try {
			$this->isMethodPost();
			$params = $this->getParameters();
			$messageParams = array(array('jobs' => $params['jobs']));
			
			$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_REQUEUE_JOBS, ProgressMapper::AUDIT_PROGRESS_REQUESTED, $messageParams); /* @var $audit \Audit\Container */
				
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, $messageParams);
			
			$this->validateMandatoryParameters($params, array('jobs'));
			$this->validateArray($params['jobs'], 'jobs');
			foreach ($params['jobs'] as $key => $jobId) {
				$this->validateInteger($jobId, "jobs[$key]");
			}
		} catch (Exception $e) {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_FAILED, array(array('errorMessage' => $e->getMessage())));
			$this->handleException($e, 'Input validation failed');
		}
	
		$requeued = array();
		$failed = array();		
		foreach ($params['jobs'] as $jobId) {
			try {
				$this->getJobqueueMapper()->requeueJob($jobId);
				$job = $this->getJobqueueMapper()->getJob($jobId);
				$requeued[] = $job;
			} catch (\Exception $e) {
				$failed[] = $jobId;
				Log::warn('Could not requeue job '. $jobId);
				Log::debug($e);
			}
		}		

		if ($failed) {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_FAILED, array($failed));
		}else {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, $messageParams);
		}
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		return array('jobs' => new JobsSet($requeued), 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet);
	}
	
	public function jobqueueJobInfoAction() {			
		try {
			$this->isMethodGet();
			$params = $this->getParameters(array(
					'id' => ''));
			$this->validateMandatoryParameters($params, array('id'));
			$this->validateInteger($params['id'], 'id');
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
	
		$job = $this->getJobqueueMapper()->getJob($params['id']);
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');

		return array('job' => $job, 'applications' => $applicationsSet, 'statuses' => $this->getDictionary()->getStatusDictionary());
	}
	
	public function jobqueueSaveRuleAction() {
		try {
			$this->isMethodPost();
			$params = $this->getParameters(array(
						'url' => '',
						'ruleId' => '-1', 
						'vars' => array(), 
						'options' => array()
					)
			);
			
			$url = trim($params['url']);//removing whitespaces
			$auditExtraData = array();
			
			if ($params['options']) {
				$auditExtraData = $auditExtraData + $params['options'];
			}
			
			if ($url) {
				$auditExtraData['url'] = $url;
			}
			
			if ($params['ruleId'] != -1) {
				$auditExtraData['ruleId'] = $params['ruleId'];// adding a ruleId in extra data only if it's not creating new rule
			}
			
			$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_SAVE_RULE, ProgressMapper::AUDIT_PROGRESS_REQUESTED, array($auditExtraData)); /* @var $audit \Audit\Container */
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, array($auditExtraData));
			
			$this->validateExclusiveMandatoryParameters($params, array('url, ruleId'));
			
			$createNewRule = true;
			if ($params['ruleId'] !== '-1') {
				$createNewRule = false;
				$this->validateInteger($params['ruleId'], 'ruleId');
			}

			if ($createNewRule) {
				$validator = new \Zend\Validator\Uri;
				if (!$validator->isValid($url)) {
				  throw new WebAPI\Exception(_t("Parameter '%s' must be a valid URL",array('url')), WebAPI\Exception::INVALID_PARAMETER);				
				} else {
					if (stripos($url, "https") === 0) {
						throw new WebAPI\Exception(_t("HTTPS schema is not supported in '%s'",array('url')), WebAPI\Exception::INVALID_PARAMETER);
					}
				}
			}
			
			$this->validateArray($params['vars'], 'vars');
						
			$options = $this->validateArray($params['options'], 'options');
			if (!isset($options['schedule'])) {
				throw new WebAPI\Exception(_t("Parameter '%s' must contain a 'schedule' element",array('options')), WebAPI\Exception::INVALID_PARAMETER);
			}
			
			foreach ($options as $idx=>$option) {
				switch ($idx) {
					case "schedule":
					case "name":
						$this->validateString($option, "options[{$idx}]");
						if (strpbrk($option, '<>')) {
							throw new WebAPI\Exception("Name contains invalid characters", WebAPI\Exception::INVALID_PARAMETER);
						}
						break;
						
					default:
						break;
				};
			}			
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
		
		if ($createNewRule) {
			$jobId = $this->getJobqueueMapper()->createRule($url, $params['vars'], $params['options']);
			if (!$jobId) {
				throw new WebAPI\Exception(_t("Failed to create rule for '%s'",array('url')), WebAPI\Exception::INTERNAL_SERVER_ERROR);
			}
			
			$jobInfo = $this->getJobqueueMapper()->getJob($jobId);
			$rule = $this->getJobqueueMapper()->getSchedulingRule($jobInfo['schedule_id']);
		} else {
			$rule = $this->getJobqueueMapper()->updateRule($params['ruleId'], $params['vars'], $params['options']);
		}		
		
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, array($auditExtraData));
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		$viewModel = new ViewModel(array('rule' => $rule, 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet));
		$viewModel->setTemplate('job-queue/web-api/jobqueue-rule-info');
		
		return $viewModel;		
	}
	
		
	public function jobqueueRunNowRuleAction() {
		try {
			$this->isMethodPost();
			$params = $this->getParameters(array(
					'ruleId' => ''
					));
			
			$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_RUN_NOW_RULE, ProgressMapper::AUDIT_PROGRESS_REQUESTED, array(array(
					'ruleId' => $params['ruleId'],
			))); /* @var $audit \Audit\Container */
				
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, array(array(
					'ruleId' => $params['ruleId'],
			)));
				
			$this->validateMandatoryParameters($params, array('ruleId'));
			$this->validateInteger($params['ruleId'], 'ruleId');						
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
	
		$job = $this->getJobqueueMapper()->runNowRule($params['ruleId']);
	Log::debug(print_r($job,true));
		
		$rule = $this->getJobqueueMapper()->getSchedulingRule($params['ruleId']);
		
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, array(array(
				'ruleId' => $params['ruleId'],
		)));
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		$viewModel = new ViewModel(array('rule' => $rule, 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet));
		$viewModel->setTemplate('job-queue/web-api/jobqueue-rule-info');
		return $viewModel;
	}
	
	public function jobqueueDeleteJobsByPredefinedFilterAction() {
		try {
			$this->isMethodPost();
			$dictionary = $this->getDictionary();
			$params = $this->getParameters(array(
					'filter' => array(),
					'filterName' => '',
			));
			$filterData = $this->validateArray($params['filter'], 'filter');
			$filterData = $this->renameFilterKeys($filterData);
			$filterName = $this->validateString($params['filterName'], 'filterName');
			
			foreach (array_keys($filterData) as $filterKey) {
				$this->validateAllowedValues($filterKey, "filter[{$filterKey}]", $dictionary->getFilterColumns());
			}			
			
			$filter = $this->getFilterObj($filterName, $filterData);
			$translator = new Translator($filter);
			
		} catch (ZendServer\Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
	
		$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_DELETE_JOBS, ProgressMapper::AUDIT_PROGRESS_REQUESTED, array($params)); /* @var $audit \Audit\Container */	
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED,  array($params));
		
		$this->getJobsMapper()->deleteJobsByFilter($translator->translate());
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, array($params));
		
		return array();
	}
	
	public function jobqueueDeleteJobsAction() {
		try {
			$this->isMethodPost();
			$params = $this->getParameters();
			$messageParams = array(array('jobs' => $params['jobs']));
			$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_DELETE_JOBS, ProgressMapper::AUDIT_PROGRESS_REQUESTED, $messageParams); /* @var $audit \Audit\Container */
			
			
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, $messageParams);
			
			$this->validateMandatoryParameters($params, array('jobs'));
			$this->validateArray($params['jobs'], 'jobs');
			foreach ($params['jobs'] as $key => $jobId) {
				$this->validateInteger($jobId, "jobs[$key]");
			}
		} catch (Exception $e) {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_FAILED, array(array('errorMessage' => $e->getMessage())));
			$this->handleException($e, 'Input validation failed');
		}
	
		$failed = array();		
		foreach ($params['jobs'] as $jobId) {
			try {
				$job = $this->getJobqueueMapper()->getJob($jobId);
				$this->getJobqueueMapper()->deleteJob($jobId);
				$deleted[] = $job;
			} catch (\Exception $e) {
				$failed[] = $jobId;
				Log::warn(_t('Could not delete job '. $jobId));
				Log::debug($e);
			}
		}		
		
		if ($failed) {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_FAILED, array($failed));
		}else {
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, $messageParams);			
		}
		
		/// clean up job status before display
		foreach ($deleted as &$job) {
			$job['status'] = \ZendJobQueue::JOB_STATUS_REMOVED;
		}

		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		return array('jobs' => new JobsSet($deleted), 'failed' => $failed, 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet);
	}
	
	public function jobqueueJobsListAction() {		
		$dictionary = $this->getDictionary();
		
		try {
			$this->isMethodGet();
			$params = $this->getParameters(array(
					'limit' => Module::config('list', 'resultsPerPage'),
					'offset' => 0,
					'order' => 'creationTime', 
					'direction' => 'DESC',
					'filter' => array(),
					'filterName' => '',
					));	
			if(isset($params['jobId']) && is_numeric($params['jobId'])){
				$jobId = $this->validateLimit($params['jobId']);
			} else {
				$jobId = null;
			}
			$limit = $this->validateLimit($params['limit']);
			$offset = $this->validateOffset($params['offset']);
			$order = $this->validateOrder($params['order'], $dictionary->getSortColumnsDictionary());
			$direction = $this->validateDirection($params['direction']);
			$filterData = $this->validateArray($params['filter'], 'filter');
			$filterData = $this->renameFilterKeys($filterData);
			$filterName = $this->validateString($params['filterName'], 'filterName');
			foreach (array_keys($filterData) as $filterKey) {
				$this->validateAllowedValues($filterKey, "filter[{$filterKey}]", $dictionary->getFilterColumns());
			}			
			
			$filter = $this->getFilterObj($filterName, $filterData);
			$translator = new Translator($filter);

			$licenseMapper = \Application\Module::serviceManager()->get('Acl\License\Mapper'); /* @var $licenseMapper \Acl\License\Mapper */
			if (isset($params['filter']['from']) && $params['filter']['from'] > 0) {
				if (!$licenseMapper->isValid($params['filter']['from']) ||  $params['filter']['from'] > $params['filter']['to']) {
					throw new LicenseException('Not valid time range');
				}
			}
			
		} catch (Exception $e) {
			$this->throwWebApiException($e, 'Input validation failed', WebAPI\Exception::INVALID_PARAMETER);
		}
		
		$jobsMapper = $this->getJobsMapper();
		$jobs = $jobsMapper->getJobsList($translator->translate(), $limit, $offset, $order, $direction, $jobId);
		//$jobs = $this->getJobqueueMapper()->getJobsList($translator->translate(), $limit, $offset, $order, $direction, $jobId);
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		return array('jobs' => $jobs, 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet);
	}

	
	public function jobqueueResumeRulesAction() {
		return $this->jobqueuePerformRulesAction("resumeRule");
	}
	
	public function jobqueueDisableRulesAction() {
		return $this->jobqueuePerformRulesAction("disableRule");
	}
	
	public function jobqueueDeleteRulesAction() {	
		try {
			$params = $this->getParameters();
			
			$audit = $this->auditMessage(Mapper::AUDIT_JOB_QUEUE_DELETE_RULES, ProgressMapper::AUDIT_PROGRESS_REQUESTED, array(array(
					'rules' => $params['rules'],
			))); /* @var $audit \Audit\Container */
			
			$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, array(array(
					'rules' => $params['rules'],
			)));
			
			$this->isMethodPost();
			$this->validateMandatoryParameters($params, array('rules'));
			$this->validateArray($params['rules'], 'rules');
			foreach ($params['rules'] as $key => $ruleId) {
				$this->validatePositiveInteger($ruleId, "rules[$key]");
			}
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
		
		$rulesChanged = array();
		Log::debug("Deleting rules " . var_export($params['rules'], true));
		foreach ($params['rules'] as $key => $ruleId) {
			$ruleChanged = $this->getJobqueueMapper()->getSchedulingRule($ruleId);
			$this->getJobsMapper()->deleteRule($ruleId);
			$rulesChanged[] = $ruleChanged;
		}
		
		foreach($rulesChanged as $id => $rule) {
			$rulesChanged[$id]['status'] = 8;//\ZendJobQueue::JOB_STATUS_REMOVED;
		}
		
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, array(array(
				'rules' => $params['rules'],
		)));
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		$viewModel = new ViewModel(array('rules' => new RulesSet($rulesChanged), 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet));
		$viewModel->setTemplate('job-queue/web-api/jobqueue-rules-list');
		return $viewModel;
	}
	
	public function jobqueueRulesListAction() {
		$this->isMethodGet();		
		$jobsMapper = $this->getJobsMapper();
		$rules = $jobsMapper->getSchedulingRules();
				
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		return array('rules' => new RulesSet($rules), 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet);
	}

	public function jobqueueRuleInfoAction() {
		try {
			$this->isMethodGet();
			$params = $this->getParameters(array(
					'id' => ''));
			$this->validateMandatoryParameters($params, array('id'));
			$this->validateInteger($params['id'], 'id');
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}
	
		$rule = $this->getJobqueueMapper()->getSchedulingRule($params['id']);
		
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
		
		return array('rule' => $rule, 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet);
	}

	public function getDictionary() { // @todo - DI
		if ($this->dictionary) {
			return $this->dictionary;
		}
	
		return $this->dictionary = $this->getLocator()->get('JobQueue\Filter\Dictionary');
	}
	
	/**
	 * @return \JobQueue\Db\Mapper
	 */
	private function getJobsMapper() {
		return $this->getLocator()->get('JobQueue\Db\Mapper');
	}
	
	private function renameFilterKeys($filterData) { // adjusting keys from the global filter widget conventions
		if (isset($filterData['to'])) {
			$filterData['executed_before'] = $filterData['to'];
			unset($filterData['to']);
		}
	
		if (isset($filterData['from'])) {
			$filterData['executed_after'] = $filterData['from'];
			unset($filterData['from']);
		}
	
		return $filterData;
	}
	
	/**
	 *
	 * @param string $filterName
	 * @return \JobQueue\Filter\Filter
	 */
	private function getFilterObj($filterName, $filterData) {
		if (!$filterName) return new Filter($filterData);
	
		$filterList = $this->getFilterMapper()->getByTypeAndName(Filter::JOB_FILTER_TYPE, $filterName); /* @var $filterList \ZendServer\Filter\FilterList */
		if (! count($filterList)) {
			throw new WebAPI\Exception(_t("Cannot find JobQueue filter '%s' in '%s' table",array($filterName, $this->getFilterMapper()->getTableName())), WebAPI\Exception::INVALID_PARAMETER);
		}
	
		return new Filter($filterData + $filterList->current()->getData()); // actual data will take precedence over dbData
	}
	
	
	private function jobqueuePerformRulesAction($action) {
		try {
			$params = $this->getParameters();
				
			switch ($action) {
				case "resumeRule":
					$auditType = Mapper::AUDIT_JOB_QUEUE_RESUME_RULES;
					break;
				case "disableRule":
					$auditType = Mapper::AUDIT_JOB_QUEUE_DISABLE_RULES;
					break;
				default:
					break;
			}
				
			$this->isMethodPost();
			$this->validateMandatoryParameters($params, array('rules'));
			$this->validateArray($params['rules'], 'rules');
			foreach ($params['rules'] as $key => $ruleId) {
				$this->validatePositiveInteger($ruleId, "rules[$key]");
			}
		} catch (Exception $e) {
			$this->handleException($e, 'Input validation failed');
		}

		$auditDetails = array(array('rules' => implode(',', $params['rules'])));
		$audit = $this->auditMessage($auditType, ProgressMapper::AUDIT_PROGRESS_REQUESTED, $auditDetails); /* @var $audit \Audit\Container */	
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_STARTED, $auditDetails);
		
		$rulesChanged = array();
		foreach ($params['rules'] as $key => $ruleId) {
			$this->getJobqueueMapper()->$action($ruleId);
			$ruleChanged = $this->getJobqueueMapper()->getSchedulingRule($ruleId);
			$rulesChanged[] = $ruleChanged;
		}	
	
		$this->auditMessageProgress($audit->getAuditId(), ProgressMapper::AUDIT_PROGRESS_ENDED_SUCCESFULLY, $auditDetails);
	
		$deploymentModel = $this->getLocator()->get('Deployment\FilteredAccessMapper'); /* @var $deploymentModel \Deployment\FilteredAccessMapper */
		$applicationsSet = $deploymentModel->getAllApplicationsInfo();
		$applicationsSet->setHydrateClass('\Deployment\Application\InfoContainer');
	
		$viewModel = new ViewModel(array('rules' => new RulesSet($rulesChanged), 'statuses' => $this->getDictionary()->getStatusDictionary(), 'priorities' => $this->getDictionary()->getPriorityDictionary(), 'applications' => $applicationsSet));
		$viewModel->setTemplate('job-queue/web-api/jobqueue-rules-list');
		return $viewModel;
	}
			
	/**
	 * @param integer $order
	 * @throws WebAPI\Exception
	 */
	private function validateOrder($order, $allowedColumns) {
		$allowedColumns = array_change_key_case($allowedColumns);
		if (! isset($allowedColumns[strtolower($order)])) {
			throw new WebAPI\Exception(_t('Parameter \'order\' must be one of %s', array(implode(', ', array_keys($allowedColumns)))), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return $order;
	}	
	
}
