<?php

namespace JobQueue\Db;

use Zend\Db\Sql\Expression;

use Zend\Json\Json;

use Audit\AuditTypeInterface;

use JobQueue\Model\JobsSet,
	JobQueue\Filter\Dictionary;

use Application\Module;

use ZendServer\Log\Log,
ZendServer\Set,
Zend\Db\TableGateway\TableGateway,
\Configuration\MapperAbstract,
Zend\Db\Sql\Select,
Zend\Db\Sql\Predicate\PredicateSet;
use ZendServer\Exception;

class Mapper extends MapperAbstract {
	
	
	private $_paramFieldToDbField = array( 
			
			"creationTime" => "creation_time",
			"startTime" => "start_time",
			"priority" => "priority",
			"status" => "status",
			"id" => "id",
			"name" => "name",
			"script" => "script",
			"application" => "app_id",
			
			
			);
	
	/**
	 * @var Dictionary
	 */
	protected $dictionary;
	
	
	public function getSchedulingRules() {
		
		try {
				
			$select = new \ZendServer\Db\Select();
			$select->from($this->getTableGateway()->getTable());
			$select->join("jobqueue_schedule", "jobqueue_schedule.id = jobqueue_job.schedule_id", array("id" => "id", "status"));
			$select->join("jobqueue_queue", "jobqueue_queue.id = jobqueue_job.queue_id", array("queue_name" => "name"));
			$select->where(array(
					"schedule_id != -1",
					"jobqueue_schedule.status != ? " => \ZendJobQueue::STATUS_REMOVED));
			
			$select->group("jobqueue_schedule.id");
			$select->columns(array(
					"name",
					"app_id" => "application_id",
					'last_run' => new \Zend\Db\Sql\Expression('MAX(start_time)'),
					"next_run" => new \Zend\Db\Sql\Expression('MAX(schedule_time)'),
					"priority",
					"persistent",
					"script",
					'type',
					'vars',
															
					));
			
			$select->order("jobqueue_job.schedule_time DESC");
			
			$result = $this->selectWith($select);
			
			// now fix the priorities
			foreach ($result as $key => $val) {
				$prt = $this->getDictionary()->dbPriorityToActualPriority((int) $val['priority']);
				$val['priority'] = $prt;
				$result[$key] = $val;
			}
			
			Log::debug("JQ rules found: " . var_export($result, true));
			
			return $result;
			
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to retrieve rules list: %s', array($e->getMessage())), null, $e);
		}
	}
	
	public function deleteRulesByAppId($appId) {
		
		$dbConnection = $this->getTableGateway()->getAdapter()->getDriver()->getConnection(); /* @var $dbConnection \Zend\Db\Adapter\Driver\ConnectionInterface */
		$dbConnection->beginTransaction();
		try {
			$dbConnection->execute("DELETE FROM jobqueue_schedule WHERE application_id = " . (int) $appId);
			$dbConnection->execute("UPDATE jobqueue_job set status = " . \ZendJobQueue::STATUS_REMOVED . " WHERE application_id = " . (int) $appId);
			$dbConnection->commit();
		} catch (\Exception $ex) {
			$dbConnection->rollback();
			Log::err($ex->getMessage());
			return false;
		}
		return true;
		
	}
	
	public function deleteRule($ruleId) {
	
		Log::debug("Deleting JQ rule $ruleId");
		
		$dbConnection = $this->getTableGateway()->getAdapter()->getDriver()->getConnection(); /* @var $dbConnection \Zend\Db\Adapter\Driver\ConnectionInterface */
		$dbConnection->beginTransaction();
		$deleted = 0;
		try {
			$deleted = $dbConnection->execute("DELETE FROM jobqueue_schedule WHERE id = " . (int) $ruleId);
			$dbConnection->commit();
		} catch (\Exception $ex) {
			$dbConnection->rollback();
			Log::err($ex->getMessage());
		}
		return $deleted;
	
	}
	
	/**
	 *
	 * @param array $filter
	 * @param integer $limit
	 * @param integer $offset
	 * @param string $sortBy
	 * @param string $direction
	 * @return \JobQueue\Model\JobsSet
	 */
	public function getJobsList($filter = array(), $limit = 0, $offset = 0, $sortBy = '', $direction = 'DESC', $singleJob = null) {		
		$total=0;
		
		try {
			
			if ($this->isSqlite()) {
				$dbConnection = $this->getTableGateway()->getAdapter()->getDriver()->getConnection(); /* @var $dbConnection \Zend\Db\Adapter\Driver\ConnectionInterface */
				$dbConnection->execute("PRAGMA read_uncommitted = 1");
			}
			
			$select = $this->createSelectFromParams($filter, $limit, $offset, $sortBy, $direction);
			Log::debug($select->getSqlString());
			
			$totalSelect = new Select();
			$totalSelect->from($this->getTableGateway()->getTable());
			$totalSelect->columns(array("count" => new \Zend\Db\Sql\Expression("COUNT(*)")));
			$totalSelect->where($select->getRawState('where'));
			$totalResult = $this->selectWith($totalSelect);
			$total = $totalResult[0]['count'];
			
			if($singleJob){
				$singleSelect = new Select();
				$singleSelect->from($this->getTableGateway()->getTable());
				$singleSelect->where('id = ' . $singleJob);
				$result = $this->selectWith($singleSelect);
				$total = 1;
			} else {
				$result = $this->selectWith($select);
			}
			
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to  retrieve job list: %s', array($e->getMessage())), null, $e);
		}
	
		$count = sizeof($result);
		Log::debug("getJobsList() returned total of {$total} jobs");
		$resultSet = new JobsSet($result);
		$resultSet->setTotal($total);
		return $resultSet;
	}
	
	public function deleteJobsByFilter(array $filter) {
		
		Log::debug("deleteJobsByFilter: " . var_export($filter, true));
		
		$where = new \Zend\Db\Sql\Where();
		if (isset($filter['priority']) && $filter['priority']) {
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression('jobqueue_job.priority IN (' . implode(',', array_values($filter['priority'])) . ')'));			
		}
		
		if (isset($filter['status']) && $filter['status']) {
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression('jobqueue_job.status IN (' . implode(',', array_values($filter['status'])) . ')'));
		} else {
			$statuses = $this->getDictionary()->getStatusColumnsDictionary();
			$removed = $statuses[Dictionary::STATUS_REMOVED];
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression("(jobqueue_job.status < $removed OR jobqueue_job.status > $removed)"));
		}
		
		if (isset($filter['app_ids']) && $filter['app_ids']) {
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression('jobqueue_job.application_id IN (' . implode(',', array_values($filter['app_ids'])) . ')'));
		}
		
		if (isset($filter['rule_ids']) && $filter['rule_ids']) {
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression('jobqueue_job.schedule_id IN (' . implode(',', $filter['rule_ids']) . ')'));
		}
		
		if (isset($filter['freeText']) && $filter['freeText']) {
			$freeText = $filter['freeText'];
			$where->addPredicate(new \Zend\Db\Sql\Predicate\Expression(array('(jobqueue_job.script LIKE ? OR jobqueue_job.name LIKE ?)' => array("%$freeText%", "%$freeText%"))));
		}
		
		if (isset($filter['executed_before']) && $filter['executed_before']) {
			$executeBefore = $filter['executed_before'];
			$where->lessThanOrEqualTo("jobqueue_job.start_time", $executeBefore);
		}
		
		if (isset($filter['executed_after']) && $filter['executed_after']) {
			$executeAfter = $filter['executed_after'];
			$where->greaterThanOrEqualTo("jobqueue_job.start_time", $executeAfter);
		}
		
		$deleted = $this->delete($where);
		
		Log::debug("deleteJobsByFilter: $deleted jobs deleted");
		
		return $deleted;
		
	}
	
	private function createSelectFromParams($filter = array(), $limit = 0, $offset = 0, $sortBy = '', $direction = 'DESC') {	
		Log::debug("received the following filter: " . var_export($filter, true));
		
		$select = new Select();
		$select->join("jobqueue_schedule", "jobqueue_schedule.id = jobqueue_job.schedule_id", array ("schedule"), Select::JOIN_LEFT);
				
		$select->columns(array (
					"id" => "id",
					"app_id" => "application_id",
					"type",
					"name",
					"node_id",
					"status",
					"priority",
					"persistent",
					"script",
					"creation_time",
					"end_time",
					"start_time",
					"variables" => "vars",
					"predecessor",
					"schedule_id",
					"http_headers",
					"output",
					"error",
					"schedule_time",
					"timeout",
					"queue_name" => new Expression('""'),
					"start_or_schedule_time" => new Expression('ifnull(jobqueue_job.start_time, jobqueue_job.schedule_time)'),			
				
				)
			);
		
		$select->from($this->getTableGateway()->getTable());
		
		$sortBy = $this->_paramFieldToDbField[$sortBy];
		
		if ($sortBy == "priority"){
			if ($direction == "DESC") { // the priorities constants in the db are order in reverse priority
				$direction = "ASC";
			} else {
				$direction = "DESC";
			}
		} else if ($sortBy == 'start_time') {
			$sortBy = "start_or_schedule_time";
		}
		
		if ($sortBy == "creation_time") {
			$sortBy = "id";
		}
		
		if (isset($filter['priority']) && $filter['priority']) {
			$select->where('jobqueue_job.priority IN (' . implode(',', array_values($filter['priority'])) . ')');			
		}
		
		if (isset($filter['status']) && $filter['status']) {
			$select->where('jobqueue_job.status IN (' . implode(',', array_values($filter['status'])) . ')');
		} else {
			$statuses = $this->getDictionary()->getStatusColumnsDictionary();
			$removed = $statuses[Dictionary::STATUS_REMOVED];
			$select->where("(jobqueue_job.status < $removed OR jobqueue_job.status > $removed)");
		}
		
		if (isset($filter['app_ids']) && $filter['app_ids']) {
			$select->where('jobqueue_job.application_id IN (' . implode(',', array_values($filter['app_ids'])) . ')');
		}
		
		if (isset($filter['rule_ids']) && $filter['rule_ids']) {
			$select->where('jobqueue_job.schedule_id IN (' . implode(',', $filter['rule_ids']) . ')');
		}
		
		if (isset($filter['freeText']) && $filter['freeText']) {
			$freeText = $filter['freeText'];
			$select->where(array('(jobqueue_job.script LIKE ? OR jobqueue_job.name LIKE ?)' => array("%$freeText%", "%$freeText%")));
		}
		
		if (isset($filter['executed_before']) && $filter['executed_before']) {
			$executeBefore = $filter['executed_before'];
			$select->where(array("(jobqueue_job.start_time <= \"{$executeBefore}\")"));
		}
		
		if (isset($filter['executed_after']) && $filter['executed_after']) {
			$executeAfter = $filter['executed_after'];
			$select->where(array("(jobqueue_job.start_time >= \"{$executeAfter}\")"));
		}
		
		$select->order(array ($sortBy => $direction));
		
		$select->limit($limit);
		$select->offset($offset);
		
		return $select;
	}
	
	protected function getDictionary() {
		if ($this->dictionary) {
			return $this->dictionary;
		}
	
		return $this->dictionary = new Dictionary();
	}
	/**
	 * @param \JobQueue\Filter\Dictionary $dictionary
	 */
	public function setDictionary($dictionary) {
		$this->dictionary = $dictionary;
	}
	
}
