<?php
namespace JobQueue\Model;

use JobQueue\Filter\Dictionary;
use ZendServer\Exception;
use ZendJobQueue, ZendJobQueueException;
use ZendServer\Log\Log;

class Mapper {

	/**
	 * @var ZendJobQueue
	 */
	protected $jobqueue;

	/**
	 * @var Dictionary
	 */
	protected $dictionary;	
	
	/**
	 * @return array
	 */
	public function getStatistics() {
		return $this->getJobqueue()->getStatistics();
	}
	
	/**
	 * 
	 * @param array $filter
	 * @param integer $limit
	 * @param integer $offset
	 * @param string $sortBy
	 * @param string $direction
	 * @return \JobQueue\Model\JobsSet
	 */
	public function getJobsList($filter = array(), $limit = 0, $offset = 0, $sortBy = 'creationTime', $direction = 'DESC', $singleJob = null) {	
		$total=0;//total would be set by getJobsList()
		$params = array(
			'sort_by' => $this->getDictionary()->getSortConstant($sortBy),
			'sort_direction' => $this->getDictionary()->getSortDirectionConstant($direction),
			'count' => (int) $limit,
			'start' => (int) $offset
		);
		
		if ($filter) {
			$params += $filter;
		}
		
		try {
			Log::debug("Calling getJobsList() with the following params: " . var_export($params, true));
			if($singleJob){
				$resultItem = $this->getJobqueue()->getJobInfo($singleJob);
				$result = array();
				$result[] = $resultItem;
			} else {
				$result = $this->getJobqueue()->getJobsList($params, $total);				
			}
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to  retrieve job list: %s', array($e->getMessage())), null, $e);
		}		
		
		$count = sizeof($result);
		Log::debug("getJobsList() returned total of {$total} jobs");
		$resultSet = new JobsSet($result);
		$resultSet->setTotal($total);
		return $resultSet;
	}
	
	/**
	 * @param string $url
	 * @param array $vars
	 * @param array $options
	 */
	public function createRule($url, array $vars, array $options) {
		try {
			$result = $this->getJobqueue()->createHttpJob($url, $vars, $options);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to create job rule: %s', array($e->getMessage())), null, $e);
		}
		
		return $result;
	}
	
	/**
	 * @param string $id
	 * @param string $vars
	 * @param string $options
	 */
	public function updateRule($id, array $vars, array $options) {
		try {
			
			$rule = $this->getSchedulingRule($id);
						
			$result = $this->getJobqueue()->updateSchedulingRule($id, $rule['script'], $vars, $options);
						
			$rule = $this->getSchedulingRule($id);
			
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to update job rule: %s', array($e->getMessage())), null, $e);
		}
	
		return $rule;
	}
	
	/**
	 * @param string $id
	 * @return array
	 */
	public function runNowRule($id) {
		try {
				
			$this->getJobqueue()->runNowSchedulingRule($id);
			$latestJob = $this->getJobsList(array(Dictionary::FILTER_COLUMN_RULE_IDS => array($id)), 1)->current();
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to execute new rule: %s', array($e->getMessage())), null, $e);
		}
	
		return $latestJob;
	}
	
	public function getJob($jobId) {
		try {
			$result = $this->getJobqueue()->getJobInfo($jobId);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to retrieve job list: %s', array($e->getMessage())), null, $e);
		}
	
		return $result;
	}
	
	/**
	 * @param integer $jobId
	 */
	public function deleteJob($jobId) {
		if ($this->getJobqueue()->getJobStatus($jobId) === false) {
			throw new Exception(_t('Failed to retrieve job'));
		}
		
		try {
			$this->getJobqueue()->removeJob($jobId);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to delete job: %s', array($e->getMessage())), null, $e);
		}
	}
	
	public function deleteJobsByFilter(array $params, $total) {
		$jobs = $this->getJobsList($params, $total);
		foreach ($jobs as $job) {
			$this->deleteJob($job['id']);
		}
	}
	
	/**
	 * @param integer $jobId
	 */
	public function requeueJob($jobId) {
		if ($this->getJobqueue()->getJobStatus($jobId) === false) {
			throw new Exception(_t('Failed to retrieve job'));
		}
		
		try {
			$this->getJobqueue()->restartJob($jobId);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to requeue job: %s', array($e->getMessage())), null, $e);
		}
	}
	
	public function createSchedulingRule(\JobQueue\Model\RecurringJob $rule) {
	    $id		= $rule->getId();
	    $script = $rule->getScript();
	    $vars	= $rule->getVars();
	    if (is_null($vars)) {
	        $vars = array();
	    }
	    $options = array();
	    $val = $rule->getName();
	    if (!empty($val)) {
	        $options['name'] = $val;
	    }
	    $val = $rule->getSchedule();
	    if (!empty($val)) {
	        $options['schedule'] = $val;
	    }
	    $q = new ZendJobQueue();
	    try {
	        $ret = $q->createHttpJob($script, $vars, $options);
	    } catch (ZendJobQueueException $e) {
	        throw new Exception($e->getMessage());
	    }
	
	    if ($ret === false) {
	        throw new Exception(_t('Failed to create a new recurring job. Make sure the Job Queue component is running and that you have the proper permissions'));
	    }
	    return true;
	}
	

	/**
	 * @param integer $ruleId
	 */
	public function resumeRule($ruleId) {
		$this->getSchedulingRule($ruleId); // validating first		
	
		try {
			$this->getJobqueue()->resumeSchedulingRule($ruleId);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to resume rule: %s', array($e->getMessage())), null, $e);
		}
	}	
	
	/**
	 * @param integer $ruleId
	 */
	public function disableRule($ruleId) {
		$this->getSchedulingRule($ruleId); // validating first
	
		try {
			$this->getJobqueue()->suspendSchedulingRule($ruleId);
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to disable rule: %s', array($e->getMessage())), null, $e);
		}
	}
	
	/**
	 * @param integer $ruleId
	 */
	public function deleteRule($ruleId) {
		$this->getSchedulingRule($ruleId); // validating first
	
		try {
			if (!$this->getJobqueue()->deleteSchedulingRule($ruleId)) {
				throw new Exception(_t('Failed to delete rule: %s', array($ruleId)));
			}
		} catch (\ZendJobQueueException $e) {
			throw new Exception(_t('Failed to delete rule: %s', array($e->getMessage())), null, $e);
		}
	}

	/**
	 * @param integer $ruleId
	 */
	public function getSchedulingRule($ruleId) {			
		if (($rule = $this->getJobqueue()->getSchedulingRule($ruleId)) === false) {
			throw new Exception(_t('Failed to retrieve rule'));
		}
	
		$rule['priority'] = \ZendJobQueue::PRIORITY_URGENT - $rule['priority'];
		return $rule;
	}		

	/**
	 * 
	 */
	public function getSchedulingRules() {			
		if (($rules = $this->getJobqueue()->getSchedulingRules()) === false) {
			throw new Exception(_t('Failed to retrieve rules'));
		}
	
		return $rules;
	}

	/**
	 * @return Boolean
	 */
	public function isJobQueueDaemonRunning() {
		return ZendJobQueue::isJobQueueDaemonRunning();
	}


	/**
	 * @return ZendJobQueue
	 * @throws Exception
	 */
	protected function getJobqueue() {
		if ($this->jobqueue) {
			return $this->jobqueue;
		}
	
		if (class_exists("ZendJobQueue")) {
			return $this->jobqueue = new ZendJobQueue();
		} else {
			throw new Exception(_t('Job Queue extension must be loaded'));
		}
		
	}
	
	protected function getDictionary() {
		if ($this->dictionary) {
			return $this->dictionary;
		}
	
		return $this->dictionary = new Dictionary();
	}
	
}