<?php
namespace ZendServer\Mvc\Controller;

use ZendServer\Mvc\Controller\ActionController,
Zend\Stdlib,
ZendServer\Mvc\Controller\Plugin\Translate,
ZendServer\Log\Log,
Zend\Validator,
WebAPI;

class WebAPIActionController extends ActionController {

    protected function validateLicenseValid() {
        $licenseMapper = $this->getZemUtilsWrapper(); /* @var $licenseMapper \Configuration\License\ZemUtilsWrapper */
        $licenseInfo = $licenseMapper->getLicenseInfo();
        if (! $licenseInfo->isLicenseOk()) {
            throw new WebAPI\Exception(_t("Zend Server license is expired or no license specified"), WebAPI\Exception::SERVER_NOT_LICENSED); 
        }
        return $licenseInfo;
    }

	/**
	 *
	 * @param integer $integer
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateInteger($integer, $parameterName) {
		if (! is_numeric($integer)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be an integer",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return intval($integer);
	}

	/**
	 *
	 * @param integer $integer
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validatePositiveInteger($integer, $parameterName) {
		$integer = $this->validateInteger($integer, $parameterName);
		if ($integer <= 0) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a positive integer",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
			
		return $integer;
	}
	
	/**
	 *
	 * @param integer $appId
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateExistingAppId($appId, $parameterName) {
	
		$deploymentMapper = $this->getLocator()->get('Deployment\Model'); /* @var $deploymentMapper \Deployment\Model */
		$apps = $deploymentMapper->getAllApplicationsInfo();
		$apps->setHydrateClass('\Deployment\Application\InfoContainer');
	
		foreach ($apps as $app) {
	
			/* @var $app \Deployment\Application\InfoContainer */
			if ($app->getApplicationId() == intval($appId) ) {
				return;
			}
		}
	
		throw new WebAPI\Exception(_t("Parameter '%s' must be an ID of an existing application",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
	}
	
	/**
	 *
	 * @param integer $serverId
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateExistingServerId($serverId, $parameterName='serverId') {
		$serverId = $this->validateInteger($serverId, $parameterName);
		
		if (! $this->getServersMapper()->isNodeIdExists($serverId)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be an existing serverId - serverId passed '%s' does not seem to be part of the configuration",array($parameterName, $serverId)), WebAPI\Exception::INVALID_PARAMETER);
		}
			
		return $serverId;
	}
	
	/**
	 * validates that the given integer is smaller than a certain passed limit
	 * @param integer $integer
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateMaxInteger($integer, $limit, $parameterName) {
		$integer = $this->validateInteger($integer, $parameterName);
		if ($integer > $limit) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be smaller than limit '%s'",array($parameterName, $limit)), WebAPI\Exception::INVALID_PARAMETER);
		}
			
		return $integer;
	}
	
	/**
	 *
	 * @param integer $percentValue
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validatePercent($percentValue, $parameterName) {
		$percentValue = $this->validateInteger($percentValue, $parameterName);
		
		if ($percentValue < 0 || $percentValue > 100) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a valid percent integer (0-100)",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
	
		return $percentValue;
	}
			
	/**
	 * 
	 * @param string $string
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateStringOrArray($param, $parameterName) {
		if (! is_string($param) && ! is_array($param)) {
			throw new WebAPI\Exception(_t("Parameter '{$parameterName}' must be a string or an array",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return $param;
	}

	/**
	 *
	 * @param array $array
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateArray($array, $parameterName) {
		if (! is_array($array)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be an array",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
	
		return $array;
	}

	/**
	 *
	 * @param array $array
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateArrayNonEmpty($array, $parameterName) {
		$this->validateArray($array, $parameterName);
		
		if (! $array) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a NON-EMPTY array, an empty array was passed", array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
	
		return $array;
	}
			
	/**
	 * 
	 * @param string $string
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateString($string, $parameterName) {
		if (! is_string($string)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a string",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return $string;
	}
	
	/**
	 *
	 * @param string $string
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateStringNonEmpty($string, $parameterName) {
		if (! is_string($string) || strlen($string) == 0) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a NON-EMPTY string",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
	
		return $string;
	}

	/**
	 *
	 * @param string $param
	 * @param string $paramName
	 * @return boolean
	 */
	protected function validateBoolean($param, $paramName) {
		$param = strtoupper($param);
		if ($param === 'TRUE') {
			return true;
		} elseif ($param === 'FALSE') {
			return false;
		} else {
			throw new WebAPI\Exception(_t("Parameter '%s' must be either 'TRUE' or 'FALSE'",array($paramName) ), WebAPI\Exception::INVALID_PARAMETER);
		}
	}

	/**
	 *
	 * @param string $host
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateHost($host, $parameterName) {
		$hostValidator = new Validator\Hostname(array('allow'=>Validator\Hostname::ALLOW_ALL));
		if (! $hostValidator->isValid($host)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a valid hostname",array($parameterName) ), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return $host;
	}
	
	/**
	 *
	 * @param string $host
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateHostWithPort($host, $parameterName) {
		
		$hostToCheck = $host;
		$parts = explode(":", $host);
		if ($parts) {
			$hostToCheck = $parts[0];
			if (isset($parts[1])) {
				$this->validatePositiveInteger($parts[1], 'port specification in ' . $parameterName);				
			}			
		} 
		
		$hostValidator = new Validator\Hostname(array('allow'=>Validator\Hostname::ALLOW_ALL));
		if (! $hostValidator->isValid($hostToCheck)) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a valid hostname",array($parameterName) ), WebAPI\Exception::INVALID_PARAMETER);
		}
	
	
		return $host;
	}
    
	/**
	 * 
	 * @param string $uri
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateUri($uri, $parameterName) {
		$validator = new \Zend\Validator\Uri;
		if (!$validator->isValid($uri)){
		    throw new WebAPI\Exception(_t("Parameter '%s' must be a valid URL",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
	    }
	    
	    return $uri;
	}
	
	/**
	 *
	 * @param string $emailAddress
	 * @param string $parameterName
	 * @throws WebAPI\Exception
	 */
	protected function validateEmailAddress($emailAddress, $parameterName) {
		$validator = new \Zend\Validator\EmailAddress();
		
		if (strstr($emailAddress, ',')) {
			$emails = explode(',', $emailAddress);
			foreach ($emails as $email) {
				if ($email) {
					$this->validateEmailAddress($email, $parameterName);
				}
			}
		} else {
			if (! $validator->isValid($emailAddress)) {
				throw new WebAPI\Exception(_t("Parameter '%s' must be a valid email address",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
			}
		}
		
		return $emailAddress;
	}

	/**
	 * @param string $value
	 * @param string $parameterName
	 * @param array $allowedValues
	 * @throws WebAPI\Exception
	 */
	protected function validateAllowedValues($value, $parameterName, array $allowedValues) {
		$lowerCaseValue = strtolower($value);
		$allowedValuesKeys = array_change_key_case(array_flip($allowedValues), CASE_LOWER); // we will now have the values as lower case keys
		
		if (!isset($allowedValuesKeys[$lowerCaseValue])) {		
			$allowedValuesStr = implode(',', $allowedValues);
			throw new WebAPI\Exception(_t("Parameter '%s' must be one of the following values: '%s'. value passed: '%s'",array($parameterName,$allowedValuesStr,$value)), WebAPI\Exception::INVALID_PARAMETER);
		}
		
		return $value;
	}	
	
	
	protected function validateValidRegex($value, $parameterName) {
		if (@preg_match("#" . $value . "#" , "") === false) {
			throw new WebAPI\Exception(_t("Parameter '%s' must be a valid regular expression",array($parameterName)), WebAPI\Exception::INVALID_PARAMETER);
		}
	}
	
	/**
	 *
	 * @param array $defaults			
	 * @return \Zend\Stdlib\Parameters
	 */
	protected function getParameters(array $defaults = array()) {
		$request = $this->getRequest(); /* @var $request \Zend\Http\PhpEnvironment\Request */
		if ($request->isGet()) {
			$parameters = $request->getQuery();
		} else {
			$parameters = $request->getPost();
		}		
		
		$parameters->fromArray(array_merge($defaults, $parameters->toArray()));
		$this->printParameters($parameters);
		return $parameters;
	}

	/**
	 *
	 * @return \MonitorUi\Model\Model
	 */
	protected function getMonitorUiModel() {
		return $this->getLocator ()->get('MonitorUi\Model\Model');
	}
	
	/**
	 *
	 * @return \MonitorUi\Model\Model
	 */
	protected function getFilteredMapper() {
		return $this->getLocator()->get('MonitorUi\Model\Model');
	}
	
	/**
	 * 
	 * @param \Zend\Stdlib\ParametersDescription $parameters
	 */
	protected function printParameters($parameters) {
		Log::debug("'{$this->getCmdName()}': the following WebAPI parameters will be used: " . trim(print_r($this->maskParameters($parameters->toArray()), true)));		
	}
	
	protected function maskParameters(array $parameters) {
		$maskedParameters = array();
		foreach ($parameters as $name=>$value) {
			if (is_array($value)) {
				$maskedParameters[$name] = $this->maskParameters($value);
				continue;
			}
			
			if (preg_match('@password@i', $name)) {
				$maskedParameters[$name] = '****';
			} else {
				$maskedParameters[$name] = $value;
			}			
		}
		
		return $maskedParameters;
	}
	
	/**
	 *
	 * @param \Zend\Stdlib\ParametersDescription $request			
	 * @param array $params of parameter names
	 * @throws WebAPI\Exception
	 */
	protected function validateMandatoryParameters(Stdlib\Parameters $requestParams, array $params) {		
		foreach ($params as $param) {
			if ((! isset ( $requestParams [$param] )) || ('' === $requestParams [$param])) {
				Log::debug(_t('This action requires the %s parameter', array($param)));
				throw new WebAPI\Exception(_t('This action requires the %s parameter', array($param)), WebAPI\Exception::MISSING_PARAMETER);
			}
		}
		
		return $params; /// returns the list of defaults!
	}
	
	/**
	 *
	 * @param \Zend\Stdlib\ParametersDescription $request
	 * @param array $params of parameter names
	 * @throws WebAPI\Exception
	 */
	protected function validateExclusiveMandatoryParameters(Stdlib\Parameters $requestParams, array $params) {
		
		foreach ($params as $param) {
			if ($requestParams [$param] !== "") {
				return;
			}
		}
		
		Log::debug(_t('This action requires one of these parameters - (%s)', array($param)));
		throw new WebAPI\Exception(_t('This action requires one of these parameters - (%s)', array($param)), WebAPI\Exception::MISSING_PARAMETER);

	}

	/**
	 * @param string $direction
	 * @throws WebAPI\Exception
	 */
	protected function validateDirection($direction) {
		return $this->validateAllowedValues($direction, 'direction', array('asc', 'desc'));
	}
	
	/**
	 * @param integer $offset
	 * @throws WebAPI\Exception
	 */
	protected function validateOffset($offset) {
		return $this->validateInteger($offset, 'offset');
	}
	
	/**
	 * @param integer $limit
	 * @throws WebAPI\Exception
	 */
	protected function validateLimit($limit) {
		return $this->validateInteger($limit, 'limit');
	}
		
	/**
	 * @throws WebAPI\Exception
	 */
	protected function isMethodPost() {
		if (! $this->isMethod('POST')) {
			throw new WebAPI\Exception(_t('This action requires a HTTP POST method'), WebAPI\Exception::UNEXPECTED_HTTP_METHOD );
		}
	}
	
	/**
	 * @throws WebAPI\Exception
	 */
	protected function isMethodGet() {
		if (! $this->isMethod ('GET')) {
			throw new WebAPI\Exception(_t('This action requires a HTTP GET method'), WebAPI\Exception::UNEXPECTED_HTTP_METHOD );
		}
	}
		
	/**
	 * @param integer $code
	 * @param string $message 
	 */
	protected function setHttpResponseCode($code, $message = '') {
		$response = $this->getResponse(); /* @var $response  \Zend\Http\Response */
		$response->setStatusCode($code);
		if ($message) {
			$response->setReasonPhrase($message);
		}
	}

	protected function getCmdName() {
		$params = $this->getEvent()->getRouteMatch()->getparams();
		return $params['action'];
	}
	
	/**
	 *
	 * @param \Exception $e
	 * @param string $msg
	 * @throws \Exception
	 */
	protected function handleException($e, $msg) {		
		Log::err("{$this->getCmdName()}: {$msg}");
		Log::debug($e);
		throw $e;
	}	

	/**
	 * 
	 * @param \Exception $e
	 * @param string $msg
	 * @param integer $code
	 * @throws WebAPI\Exception
	 */
	protected function throwWebApiException($e, $msg, $code) {
		Log::err("{$this->getCmdName()}: {$msg}");
		Log::debug($e);		
		throw new WebAPI\Exception(_t("%s: '%s'", array($msg, $e->getMessage())), $code);
	}
		
	/**
	 * @param string $method
	 *			one of POST,GET
	 * @return boolean
	 */
	private function isMethod($method) {
		$request = $this->getRequest (); /* @var $request \Zend\Http\Request */
		return strcasecmp ( $request->getMethod (), $method ) === 0;
	}
	
}

