<?php

abstract class ZServer_Cli_Tool_Command_Manage_Abstract extends ZServer_Cli_Tool_Command_Abstract
{
	const EXIT_AUTHERROR  = 31;
	const EXIT_ZSCMLOCKED = 32;
	
	protected $_zsApiClient = null;
	
	/**
	 * Server URL
	 *
	 * @var Zend_Uri_Http
	 */
	protected $_serverUrl = null;

	/**
	 * Wrapper around the run() method that handles service exceptions
	 * 
	 * @see ZServer_Cli_Tool_Command_Abstract::run()
	 */
	public function run()
	{
		try {
			$this->_run();
			
		} catch(ZendX_Service_ZendServer_Exception $ex) {
			$this->_tool->output()->debug((string) $ex);
			
			switch ($ex->getZendServerErrorCode()) {				
				case 'authError':
					throw new ZServer_Cli_Tool_Exception("authentication error. check your API key details", self::EXIT_AUTHERROR);
					break;
					
				case null:
					throw new ZServer_Cli_Tool_Command_Exception("API error: {$ex->getMessage()}", ZServer_Cli_Tool::EXIT_UNKNOWN);
					break;
					
				default:
					//$this->_tool->output()->debug("WebAPI error code: " . $ex->getZendServerErrorCode() . PHP_EOL);
					throw new ZServer_Cli_Tool_Exception('failed accessing Zend Server through the Web API' . PHP_EOL . 
					'WebAPI error message : ' . $ex->getMessage() . PHP_EOL . 
					'WebAPI error code    : ' . $ex->getZendServerErrorCode(), ZServer_Cli_Tool::EXIT_UNKNOWN);
					break;
			}
		}
	}
	
	/**
	 * Actual execution method, called by run()
	 * 
	 * @see ZServer_Cli_Tool_Command_Abstract::run()
	 */
	abstract protected function _run();
	
	/**
	 * Get the server URL
	 * 
	 * @return Zend_Uri_Http
	 */
	protected function _getServerUrl()
	{
		if (! $this->_serverUrl) { 
			$url = $this->_tool->getArguments()->U;
			if (! $url) {
				if (getenv('ZS_API_SERVER_URL')) { 
					$url = getenv('ZS_API_SERVER_URL');
				} else { 					
					$zscm = (boolean) preg_grep('/PRODUCT_NAME="Zend Server Cluster Manager"/', @file('/etc/zce.rc'));// TODO - will not work on windows
					$zscm === true ? $path = 'ZendServerManager' : $path = 'ZendServer';
					$url = 'http://localhost:10081/' . $path;
				}
			}
			
			try {
				$this->_serverUrl = Zend_Uri_Http::factory($url);
				
			} catch (Zend_Uri_Exception $ex) {
				$this->_tool->output()->debug((string) $ex);
				throw new ZServer_Cli_Tool_Exception("Invalid server URL '$url': " . $ex->getMessage(), ZServer_Cli_Tool::EXIT_INVALID_ARG); 
			}
		}
		
		return $this->_serverUrl;
	}
	 
	/**
	 * Preapre the userParams as an assoicative array
	 * 
	 * @param string $params - a string in the form of "a=1 b=2"
	 * @return array 
	 */   
	protected function _getUsersParams($params) {
		$user_params = array();
		$param_couples = explode(' ', $params);
		
		foreach ($param_couples as $couple) {
			@list($key, $value) = explode('=', $couple);
			if ($value === null) throw new ZServer_Cli_Tool_Command_Exception('userParams not passed properly - should be of the form of "a=1 b=2"', ZServer_Cli_Tool::EXIT_INVALID_ARG);
			$user_params[$key] = $value;
		}
		
		return $user_params;
	}
		
	/** 
	 * Get the ZS API client used by this object
	 * 
	 * @return ZendX_Service_ZendServer 
	 */
	protected function _getZsApiClient()
	{
		$apiKeyName = $this->_tool->getArguments()->N;
		$apiKey = $this->_tool->getArguments()->K;
		if (!$this->_name==="bootstrap-single-server") {
			if (! $apiKeyName) {
				if (getenv('ZS_API_KEY_NAME')) { 
					$apiKeyName = getenv('ZS_API_KEY_NAME');
				} else {
					throw new ZServer_Cli_Tool_Exception("API key name not specified", ZServer_Cli_Tool::EXIT_INVALID_ARG);
				}
			}
			
			if (! $apiKey) {
				if (getenv('ZS_API_KEY')) { 
					$apiKey = getenv('ZS_API_KEY');
				}  else {
					throw new ZServer_Cli_Tool_Exception("API secret key not specified", ZServer_Cli_Tool::EXIT_INVALID_ARG);
				}
			}
		}
		
		if (! $this->_zsApiClient) { 
			$this->_zsApiClient = new ZendX_Service_ZendServer(
				$this->_getServerUrl(), $apiKeyName, $apiKey 
			);
		}
		
		return $this->_zsApiClient; 
	}
	
	/**
	 * Print server information
	 * 
	 * @param ZendX_Service_ZendServer_ServerInfo $serverInfo
	 */
	protected function _printServerInfo(ZendX_Service_ZendServer_ServerInfo $serverInfo)
	{
		$output = $serverInfo->getId() . "\t" . 
				  $serverInfo->getName() . "\t" . 
				  $serverInfo->getAddress() . "\t" . 
				  $serverInfo->getStatus() . PHP_EOL;
				  
		foreach($serverInfo->getMessages() as $message) { /* @var $message ZendX_Service_ZendServer_Message */
			$output .= "\t" . $message->getLevel() . "\t" . $message->getMessage() . PHP_EOL;
		}
		
		$this->_tool->output()->write($output);
	}
	
	/**
	 * Print application information
	 * 
	 * @param ZendX_Service_ZendServer_ApplicationInfo $applicationInfo
	 */
	protected function _printApplicationInfo(ZendX_Service_ZendServer_ApplicationInfo $applicationInfo)
	{
		$output = 	PHP_EOL . "APPINFO" . "\t" . 
					$applicationInfo->getId() . "\t" . 
					$applicationInfo->getBaseUrl() . "\t" . 
					$applicationInfo->getAppName() . "\t" . 
					$applicationInfo->getUserAppName() . "\t" . 
					$applicationInfo->getInstalledLocation() . "\t" . 				
					$applicationInfo->getStatus() . "\t" .  PHP_EOL;


		foreach($applicationInfo->getServers() as $app_server) { /* @var $app_server ZendX_Service_ZendServer_ApplicationServer */
			$output .= "\t" . 'SERVERAPPINFO' . "\t" . $app_server->getId() . "\t" . $app_server->getDeployedVersion() . "\t" . $app_server->getStatus() . PHP_EOL;
		}

		foreach($applicationInfo->getDeployedVersions() as $DeployedVersions) {
			$output .= "\t" . 'DEPLOYEDVERSION' . "\t" . $DeployedVersions . PHP_EOL;
		}			

		foreach($applicationInfo->getMessages() as $message) { /* @var $message ZendX_Service_ZendServer_Message */
			$output .= "\t" . 'APPMESSAGE' . "\t" . $message->getLevel() . "\t" . $message->getMessage() . PHP_EOL;
		}			

		
		$this->_tool->output()->write($output);
	}
	
	protected function _printLibraryInfo(ZendX_Service_ZendServer_LibraryInfo $libraryInfo)
	{
		//var_dump($libraryInfo);
		$output = 	PHP_EOL . "APPINFO" . "\t" .
					$libraryInfo->getId() . "\t" .
					$libraryInfo->getAppName() . "\t" .
					$libraryInfo->getStatus() . "\t" .  PHP_EOL;
		
		foreach($libraryInfo->getVersions() as $version) {  /* @var $version ZendX_Service_ZendServer_LibraryVersion */
			//var_dump($version);
			$output .= "\t" . 'APPVERSION'  . "\t" . $version->getLibraryVersionId() . "\t" . $version->getVersion() . "\t" . $version->getStatus() . "\t" . $version->getInstalledLocation() . PHP_EOL;
		}
		$this->_tool->output()->write($output);
	}
	   
	/**
	 * Re-check server status every few seconds until it's status changes
	 * 
	 * @param  ZendX_Service_ZendServer_ServerInfo $serverInfo
	 * @return ZendX_Service_ZendServer_ServerInfo
	 */
	protected function _waitUntilServerStatusChange(ZendX_Service_ZendServer_ServerInfo $serverInfo)
	{
		$origStatus = $serverInfo->getStatus();
		$serverId = $serverInfo->getId();
		
		do {
			sleep(5);
			$servers = $this->_getZsApiClient()->clusterGetServerStatus(array($serverId));
			if (isset($servers[$serverId])) { 
				$serverInfo = $servers[$serverId];
			} else { 
				break;
			}
			
		} while ($serverInfo->getStatus() == $origStatus);
			
		return $serverInfo;
	}
}
