<?php

class ZServer_Cli_Tool_Command_Setup_CreateClusterDb extends ZServer_Cli_Tool_Command_Abstract
{
    const EXIT_DB_ERROR = 51;
    
    const DB_ADAPTER = 'Pdo_Mysql';
    
    protected $_description = 'create the ZSCM monitor MySQL database';
    
    protected $_usageMessage = '-H <db host> -u <db user> -p <db password> [options]';

    /* TODO: add option to create DB if root user / pass is specified */
    protected $_args = array(
        'H|db-host=a'    => 'database hostname',
        'u|db-user=a'    => 'database username', 
        'p|db-pass=a'    => 'database password',
        'n|db-name=a'    => 'database schema name (default: zend_monitor)',
    	'P|db-port=a'    => 'database port (default: 3306)',
        'C|create-db'    => 'create the DB schema fron scratch. Provided user is assumed to be root',
        'L|create-local' => 'when connecting as root, connect to localhost instead of the provided hostname'
    );
    
    public function run()
    {
        if (ZEND_SERVER_APP_VERSION != 'EE') { 
            throw new ZServer_Cli_Tool_Command_Exception("this command can only be used on a ZSCM installation", ZServer_Cli_Tool::EXIT_UNKNOWN);
        }
        
        $dbHost = $this->_tool->getArguments()->H;
        $dbUser = $this->_tool->getArguments()->u;
        $dbPass = $this->_tool->getArguments()->p;
        $dbPort = $this->_tool->getArguments()->P;
        $dbName = $this->_tool->getArguments()->n;
        
        if (! $dbHost) { 
            throw new ZServer_Cli_Tool_Command_Exception("missing database hostname argument", ZServer_Cli_Tool::EXIT_INVALID_ARG);
        }
        
        // Check that the provided hostname is not 'localhost' or equivalent
        if (substr(gethostbyname($dbHost), 0, 4) == '127.') {
            throw new ZServer_Cli_Tool_Command_Exception("specified database hostname cannot resolve to localhost", ZServer_Cli_Tool::EXIT_INVALID_ARG);
        }
        
        if (! $dbUser) { 
            throw new ZServer_Cli_Tool_Command_Exception("missing database username argument", ZServer_Cli_Tool::EXIT_INVALID_ARG);
        }
        
        if (! $dbPass) $dbPass = ''; // empty password?
        
        if (! $dbName) {
            $dbName = DatabaseManager_Mapper::ZEND_MONITOR_SCHEMA;
        }
        
        $dbPort = (int) $dbPort;
        if (! $dbPort) $dbPort = 3306; 
        
        if ($this->_tool->getArguments()->C) {
            // Create schema
            $creds = $this->_createMonitorDb($dbHost, $dbPort, $dbUser, $dbPass, $dbName, $this->_tool->getArguments()->L);
            $dbUser = $creds['username'];
            $dbPass = $creds['password'];
            
        } else {
            // Populate an existing schema
            $this->_populateMonitorDb($dbHost, $dbPort, $dbUser, $dbPass, $dbName);
        }
        
        // Save settings
        $this->_saveIniDbSettings($dbHost, $dbPort, $dbUser, $dbPass, $dbName);
        
        // restart jqd
        $this->_restartJqd();
    }
    
    /**
     * Create an empty database using root credentials. 
     * 
     * $dbUser and $dbPassword will be modified to contain the new unprivileged
     * user's credentials
     * 
     * If $connectLocal is TRUE, will use 'localhost' connection to create the
     * DB. The INI files will still contain the external host name. This is 
     * useful when root has no privileges to connect from other servers.
     *  
     * @param string  $dbHost
     * @param integer $dbPort
     * @param string  $dbUser
     * @param string  $dbPass
     * @param string  $dbName
     * @param boolean $connectLocal
     * @return array Created user credentials array ('username' => ..., 'password' => ...)
     */
    protected function _createMonitorDb($dbHost, $dbPort, $dbUser, $dbPass, $dbName, $connectLocal)
    {
        if ($connectLocal) {
            $dbHost = '127.0.0.1';
        }
        
        $dbManager = new DatabaseManager_Mapper(
            $dbHost,
            $dbPort,
            $dbUser,
            $dbPass,
            $dbName
        );
        
        $monitorModel = new ZwasComponents_ZSM_Monitor_Model();
        
        try {
            $monitorModel->createUserAndSchema($dbManager);
        } catch (Zwas_Exception $ex) {  	
            $this->_tool->output()->debug($ex->getMessageObject()->__toString());
            throw new ZServer_Cli_Tool_Command_Exception("unable to create DB: {$ex->getMessageObject()->__toString()}", self::EXIT_DB_ERROR);
        }
        
        return $dbManager->getCreatedUserCredentials();
    }
    
    /**
     * Populate the monitor DB empty schema
     * 
     * @param string  $dbHost
     * @param integer $dbPort
     * @param string  $dbUser
     * @param string  $dbPass
     * @param string  $dbName
     */
    protected function _populateMonitorDb($dbHost, $dbPort, $dbUser, $dbPass, $dbName)
    {
        $dbManager = new DatabaseManager_Mapper(
            $dbHost,
            $dbPort,
            $dbUser,
            $dbPass,
            $dbName
        );
        
        $monitorModel = new ZwasComponents_ZSM_Monitor_Model();
        
        try {
            $monitorModel->createSchemaOnly($dbManager);
        } catch (Zwas_Exception $ex) { 
            $this->_tool->output()->debug($ex->getMessageObject()->__toString());
            throw new ZServer_Cli_Tool_Command_Exception("unable to create DB: {$ex->getMessageObject()->__toString()}", self::EXIT_DB_ERROR);
        }
    }
    
    /**
     * Save the new monitor DB configuration to the INI filesDatabaseManager_Model
     * 
     * @param string  $dbHost
     * @param integer $dbPort
     * @param string  $dbUser
     * @param string  $dbPass
     * @param string  $dbName
     */
    protected function _saveIniDbSettings($dbHost, $dbPort, $dbUser, $dbPass, $dbName)
    {
		//$monitorModel = new ZwasComponents_ZSM_Monitor_Model();
		$model = new DatabaseManager_Model; // will also set zdd.ini 
		Zend_Registry::set('deploymentSupported', true); // that supportedOnCurrentEnviroment() will eval to true
		$model->saveDbCredentials($dbHost, $dbPort, $dbUser, $dbPass, $dbName);
    }
    
    /**
     * restart the JQD daemon/service
     */
    protected function _restartJqd() {
    	$jobQueueDaemon = new ZwasComponents_JobQueue_Daemon_Model();
    	$stopResult = $jobQueueDaemon->stopService();
    	$startResult = $jobQueueDaemon->startService();    	
    }
}